/*****************************************************************************
                                RW_FITS

Author:  Charles Dyer, Chris Burns

A package of functions to ease the use of FITS files

variable type:
typedef struct
   {
   int simple;          whether or not it is simple.  It had better be! 
   int bitpix;          8(char), 16(int), 32(long int), or -32(float)  
   int n_axis;          number of axes (up to 4)
   int naxis[];         number of pixels per axis
   int blank;           value of a blank pixel (int type only)
   float crpix[];       center pixel value
   float datamax;       maximum of data[]
   float datamin;       minimum of data[]
   float crval[];       offset in coordinates
   float cdelt[];       width of a pixel in sky coordinates
   float cd[];          wcs rotation/shear/scale matrix
	float pc[];          wcs rotation/shear matrix (no scaling:  use cdelt[])
   char object[80];     object name
   other_node *other;   linked list of "other" header cards (see functions)
   float *data;         pointer to data array
   } fits;

procedures:

   int read_fits_file (char *filename, fits *)
	   A wrapper for read_fits so that you can just pass it the name of a
		file.  I use this for Python, which doesn't really handle FILE
		structs very well.
	
	int write_fits_file (char *filename, fits *)
		A wrapper for write_header and write_data.  This will take the fits
		data in fits * and write it to the file *filename.

   int read_fits(FILE *file, fits *input)
      Read in a fits file pointed to by *file and fill in the fits variable
      input.  This is a wrapper for read_header, alloc_fits and read_data,
      so you don't have to worry about memory allocation for the *data;  it
      just has to be a valid pointer.

   int cp_header(fits *fits1, fits *fits2)
      Read the header from file1 and output it to file2.  Useful for copying
      an entire header, as is, while bypassing the read_fits routine.

   int write_header(FILE *file, fits output)
      Take the header in the fits structure output and write it out to the
      file pointed to by *file.

   int write_data(FILE *file, fits output)
      Write the data located in output to the file pointed to by *file.

	void print_header (fits)
	   prints out a summary of the fits data.

	cp_data( fits, fits *)
	   copy the data from the first fits to the second.  The fits structures
		have to have the same size and the memory has to be allocated.

   float read_fits_pixel(int)
		get the value at a pixel.  The parameter is the index of the pixel.

	int search_key (fits, char *pattern, char *buffer)
	   search for the pattern *pattern and return the first fits header line
		that matches.

	int change_key (fits, char *key, char *value)
	   Change the header key-value pair.  This is just strings, so there is
		no checking to make sure that you are following the standard FITS
		conventions.  Beware.

	int alloc_fits( fits *)
	   allocate the data array in the fits structure.  It uses the naxis
		and naxis[] arrays to make the appropriate size, so make sure these
		are set before calling this function.

	int get_ndata(fits)
	   gets the total size of the data array:  naxis[0]*naxis[1]*...*naxis[n]

   
*/
#ifdef __cplusplus
extern "C" {
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "rw_fits.h"

/* function prototypes */
int value( fits *, char *, short int, char * );
struct other_node *alloc_node();
int read_header(FILE *, fits *);
int write_header(FILE *, fits);
int cp_header(fits , fits *);
int must_swap();
void l_swap(unsigned char *);
void s_swap(unsigned char *);
int alloc_fits ( fits *);
void read_data(FILE *, fits *);
int write_data(FILE *, fits );
int read_fits ( FILE *, fits *);
int read_fits_file (char *, fits *);
int write_fits_file (char *, fits);
void print_header( fits );
int cp_data( fits , fits *);
int init_fits(fits *);
long int get_ndata(fits);

int fdebug = 0;

/* Routine to add a key/value pair for the "other" structure in the FITS
   type */
int addkey( fits *input, char *key, short int eq, char *value ) {

   struct other_node *node, *node2;

   /* node = alloc_node(); */
   if ( ! (node = (struct other_node *) malloc (sizeof(struct other_node)))) { 
      fprintf(stderr, "Bad allocation for other_node\n");
      exit(1);
   }
   node->key[8] = 0;
   node->value[70] = 0;

   strncpy(node->key, key, 8);
   node->eq = eq;
   strncpy(node->value,value, 70);

   /* find the end of the linked list and add the new node */
   if ( input->other == (struct other_node *) NULL ) {
      input->other = node;
      node->next = (struct other_node *) NULL;
   }
   else {
      for( node2 = input->other ; node2 != (struct other_node *) NULL ; 
           node2 = node2->next ) {
         if (node2->next == (struct other_node *) NULL) {
            node2->next = node;
            node->next = (struct other_node *) NULL;
            break;
         }
      }
   }
   return(0);
}

/* My own allocator to hand out other_nodes */
/* Depricated!  The use of object oriented programming defeats this lovely
   trick.  You can't dealloc a small part of the block.  Back to old malloc() */
struct other_node *alloc_node() {
   static struct other_node *free;      /* points to the next free node */
   static struct other_node *temp;      /* temporary pointer */
   struct other_node *memory_chunk;     /* location of the memory chunk */
   struct other_node *current;          /* used for initialization */
   int i;

   if (free == (struct other_node *) NULL ) {   /* 1st call or no more left */
      /* Allocate a large (36) block of other_nodes */

      if ( ! (memory_chunk = (struct other_node *) calloc (36,
           sizeof(struct other_node)))) {
         fprintf(stderr, 
            "Error!  Bad memeory allocation in alloc_node. Exit!\n");
         exit(1);
      }
      else {
         /* Now we make a linked list of these nodes (easier to hand them 
            off */
         for (current = memory_chunk, i = 0; i < 36 - 1 ; ++i) {
            current->next = current + 1;
            current = current->next;
         }
         current->next = (struct other_node *) NULL;
         free = memory_chunk;
      }
   }

   /* temp is the pointer we return.  Increment the free pointer to the
      next in the chunk.  Break the temp pointer away from the chunk and
      finally return it */
   temp = free;
   free = free->next;
   temp->next = (struct other_node *) NULL;
   return(temp);
}

/* Routine to search through the other_nodes for a pattern.  The entire
   FITS record is written to buffer, to do with as one wishes.  Returns 0 if 
   not found, 1 if found.  buffer must be at least 80 characters long. */

int search_key(fits var, char *pattern, char *buffer) {

   struct other_node *node;
   char *where;

   if (var.other == (struct other_node *) NULL ) {
      return(-1);
   }
   for ( node = var.other ; node != (struct other_node *) NULL ; 
         node = node->next ) {
      where = strstr(node->key, pattern);
      /*fprintf(stderr, "where = %s\n", where);*/
      if (! where) where = strstr(node->value, pattern);
      if (where) {
         if ( buffer != (char *) NULL) {
            if (node->eq) {
               sprintf(buffer, "%s=%s", node->key, node->value);
            } else {
               sprintf(buffer, "%s%s", node->key, node->value);
            }
         }
         return(0);
      }
   }
   return(-2);
}

/* Routine to search through the other_nodes for a key.  The value is
 * changed to the new value input */

int change_key(fits var, char *pattern, char *value) {

   struct other_node *node;
   char *where;

   if (var.other == (struct other_node *) NULL ) {
      return(-1);
   }
   for ( node = var.other ; node != (struct other_node *) NULL ; 
         node = node->next ) {
      where = strstr(node->key, pattern);
      if (where) {
         if (!node->eq) return(-3);   /* a comment tag */
         strncpy(node->value,value, 70);
         return(0);
      }
   }
   return(-2);
}


fits *new_fits() {
   fits *new_struct;

   if ( ! (new_struct = (fits *)malloc(sizeof(fits)))) { 
      fprintf(stderr, "Error, memory allocation for new fits struct\n");
      exit(1);
   }
   init_fits(new_struct);
   return new_struct;
}

void free_other_nodes (fits *fitsvar) {
   struct other_node *here, *there;

   if ( fitsvar->other != (struct other_node *)NULL ) {
      if (fitsvar->other->next != (struct other_node *)NULL) {
         here = fitsvar->other;
         there = here->next;
         while (there->next != (struct other_node *) NULL) {
            free(here);
            here = there;
            there = there->next;
         }
         free(there);
         free(here);
         fitsvar->other = (struct other_node *) NULL;
      }
      else {
         free(fitsvar->other);
         fitsvar->other = (struct other_node *) NULL;
      }
   }
}

void free_fits_struct(fits *fitsvar) {

   free_other_nodes(fitsvar);
   if (fitsvar->data != (float *)NULL) {
      free(fitsvar->data);
      fitsvar->data = (float *)NULL;
   }
   if(fitsvar->naxis != (int *) NULL) free(fitsvar->naxis);
   if(fitsvar->crval != (float *) NULL) free(fitsvar->crval);
   if(fitsvar->cdelt != (float *) NULL) free(fitsvar->cdelt);
   if(fitsvar->crpix != (float *) NULL) free(fitsvar->crpix);
   if(fitsvar->cd != (float *) NULL) free(fitsvar->cd);
   if(fitsvar->pc != (float *) NULL) free(fitsvar->pc);
   free(fitsvar);
}

/* initialize the members of the fits header */
int init_fits(fits *var) {
   var->simple = 1;
   var->bitpix = 0;
   var->n_axis = -1;
   var->blank = 0;
   var->useblank = 0;
   var->datamax = 0;
   var->datamin = 0;
   sprintf(var->object, "''                                                                             "); 
   var->other = (struct other_node *) NULL;
   var->naxis = (int *) NULL;
   /*var->naxis[0] = 0;var->naxis[1] = 0;var->naxis[2] = 0;var->naxis[3] = 0;*/
   /*var->crpix[0] = 0;var->crpix[1] = 0;var->crpix[2] = 0;var->crpix[3] = 0;*/
   /*var->crval[0] = 0;var->crval[1] = 0;var->crval[2] = 0;var->crval[3] = 0;*/
   /*var->cdelt[0] = 0;var->cdelt[1] = 0;var->cdelt[2] = 0;var->cdelt[3] = 0;*/
   var->data = (float *) NULL;
   return(0);
}
   
/* 
   Routine to read in the header for a fits file pointed to by *fp and store
   that information into the fits struct variable, and return the offset to
   the end of the header.
*/
int read_header(FILE *fp, fits *var)
{
   char fcard[80]; /* basic FITS card, 36 to a block of 2880 bytes */
	int ai=0, aj=0;
   int end = 0; /* assume the worst! */
   int k = 0, i;
   int total = 0;
   char *token;
   var->datamax = var->datamin = 0;    /* 0 will ==> not specified */
   var->blank = 0;
   var->useblank = 0;                    /* 0 will ==> not specified */
   var->other = (struct other_node *) NULL;
   while( fread(fcard,1,80,fp) == 80 ) {
      ++k;
      if (strncasecmp(fcard+8, "= ", 2) == 0 ) {
         addkey(var,fcard, 1, fcard+10);
      }
      else {
         addkey(var, fcard, 0, fcard+8);
      }
      if( strncasecmp(fcard,"BITPIX",6) == 0 ) { var->bitpix = atol(fcard+9); }
      else
      if( strncasecmp(fcard,"NAXIS ",6) == 0 ) { 
         var->n_axis = atol(fcard+9); 
         var->naxis = (int *) malloc ( var->n_axis * sizeof(int));
         var->crval = (float *) malloc ( var->n_axis * sizeof(float));
         var->cdelt = (float *) malloc ( var->n_axis * sizeof(float));
         var->crpix = (float *) malloc ( var->n_axis * sizeof(float));
         var->cd = (float *) malloc ( var->n_axis*var->n_axis * sizeof(float));
         var->pc = (float *) malloc ( var->n_axis*var->n_axis * sizeof(float));
         for (i = 0 ; i < var->n_axis ; ++i) {
            var->naxis[i] = -1;   /* invalid entry */
            var->crval[i] = 0;
            var->cdelt[i] = 0;
            var->crpix[i] = 0;
         }
			for (i = 0 ; i < var->n_axis*var->n_axis ; ++i) {
				var->cd[i] = 0;
				if (i % var->n_axis == i / var->n_axis) {
					var->pc[i] = 1.0;
				} else {
					var->pc[i] = 0.0;
				}
			}
      }
      else
      if( strncasecmp(fcard,"NAXIS",5) == 0 ) {
         int axis = 0;
         sscanf(fcard, "NAXIS%d", &axis);
         axis = axis - 1;    /* NAXIS tags start at 1 */
         if (axis < 0 || axis > var->n_axis - 1) {
            fprintf(stderr, "Warning!  NAXIS%d is invalid for %d dimensions\n",
                    axis+1, var->n_axis);
         } else {
            var->naxis[axis] = atol(fcard+9); 
         }
      }
      else
      if( strncasecmp(fcard,"SIMPLE",5) == 0 ) { var->simple = 1; }
      else
      if( strncasecmp(fcard,"END",3) == 0 ) { end = 1; break; }
      else
      if( strncasecmp(fcard,"DATAMAX",7) == 0 ){ var->datamax = atof(fcard+9); }
      else
      if( strncasecmp(fcard,"DATAMIN",7) == 0 ){ var->datamin = atof(fcard+9); }
      else
      if( strncasecmp(fcard,"BLANK",5) == 0 ) { 
         var->blank = atol(fcard+9);
         var->useblank = 1;
      }
      else
      if( strncasecmp(fcard,"CRVAL",5) == 0 ) {
         int axis = 0;
         sscanf(fcard, "CRVAL%d", &axis);
         axis = axis - 1;    /* NAXIS tags start at 1 */
         if (axis < 0 || axis > var->n_axis - 1) {
            fprintf(stderr, "Warning!  CRVAL%d is invalid for %d dimensions\n",
                    axis+1, var->n_axis);
         } else {
            var->crval[axis] = atof(fcard+9); 
         }
      }
      else
      if( strncasecmp(fcard,"CDELT",5) == 0 ) {
         int axis = 0;
         sscanf(fcard, "CDELT%d", &axis);
         axis = axis - 1;    /* NAXIS tags start at 1 */
         if (axis < 0 || axis > var->n_axis - 1) {
            fprintf(stderr, "Warning!  CDELT%d is invalid for %d dimensions\n",
                    axis+1, var->n_axis);
         } else {
            var->cdelt[axis] = atof(fcard+9); 
         }
      }
      else
      if( strncasecmp(fcard,"CRPIX",5) == 0 ) {
         int axis = 0;
         sscanf(fcard, "CRPIX%d", &axis);
         axis = axis - 1;    /* NAXIS tags start at 1 */
         if (axis < 0 || axis > var->n_axis - 1) {
            fprintf(stderr, "Warning!  CRPIX%d is invalid for %d dimensions\n",
                    axis+1, var->n_axis);
         } else {
            var->crpix[axis] = atof(fcard+9); 
         }
      }
      else
      if( strncasecmp(fcard,"CD",2) == 0 && 
		    sscanf(fcard, "CD%d_%d", &ai, &aj) > 0 ) {
         ai = ai - 1;  aj = aj - 1;  /* NAXIS tags start at 1 */
         if (ai < 0 || ai > var->n_axis - 1 || aj < 0 || aj > var->n_axis - 1) {
            fprintf(stderr, "Warning!  CD%d_%d is invalid for %d dimensions\n",
                    ai+1, aj+1, var->n_axis);
         } else {
            var->cd[ai + aj*var->n_axis] = atof(fcard+9); 
         }
      }
		else
      if( strncasecmp(fcard,"PC",2) == 0 &&
          sscanf(fcard, "PC%d_%d", &ai, &aj) > 0 ) {
         ai = ai - 1;  aj = aj - 1;  /* NAXIS tags start at 1 */
         if (ai < 0 || ai > var->n_axis - 1 || aj < 0 || aj > var->n_axis - 1) {
            fprintf(stderr, "Warning!  PC%d_%d is invalid for %d dimensions\n",
                    ai+1, aj+1, var->n_axis);
         } else {
            var->pc[ai + aj*var->n_axis] = atof(fcard+9); 
         }
      }
      else
      if( strncasecmp(fcard,"OBJECT",6) == 0 ) {
         token = strtok(fcard+9, "'");
         if (token) token = strtok(NULL, "'");
         if (token) {
            strcpy(var->object, token);
         }
         else {
            strcpy(var->object, "No Object Tag");
         }
         /* sscanf(fcard+9, "'%s'", var->object); */
      }
    }

    if( !var->simple ) {
       return(-1);
    }

    if( !end ) { 
       return(-2); 
    }
    /* Now round up to end of header block containing END card */
    total = k;
    if (k % 36 > 0) {
       for( k = k % 36 ; (k < 36) && (fread(fcard,1,80,fp) == 80) ; ++k ) {
          ++total;
       }
    }
    /* return the total number of fits header cards read */
    return(total);
}

/* Routine to write in the header for a fits file pointed to by *fp and 
   round up to the last fits 2880 block.   Info is taken from the fits-type
   argument.
*/
int write_header(FILE *fp, fits var)
{
   char fcard[81]; /* basic FITS card, 36 to a block of 2880 bytes */
   struct other_node *node;
   int i;
   int k=0;

   for ( node = var.other ; node != (struct other_node *) NULL ; 
         node=node->next) {
      if (node->eq == 0) {
         sprintf(fcard, "%-8s  %-70s", node->key, node->value);
      }
      else {
         sprintf(fcard, "%-8s= %-70s", node->key, node->value);
      }
      fwrite(fcard,1,80,fp); ++k;
   }

   /* sprintf(fcard, "%-80s", "END");
   fwrite(fcard,1,80,fp); ++k; */

   /* Now round up to end of header block containing END card */
   k = k*80;    /* actual number of bytes written */

   if (k % 2880 > 0) {
      for( k = k % 2880 ; (k < 2880) ; ++k ) {
         fwrite(" ", 1, 1, fp);
      }
   }
   return(k);
}

/*  A routine to copy the header from one fits variable (pointed to by *src) to 
    another (pointed to by *dest).  */

int cp_header(fits src, fits *dest)
{
   int i = 0;
   struct other_node *node;
   
   /* copy everything over */
   dest->simple = src.simple;
   dest->bitpix = src.bitpix;
   dest->n_axis = src.n_axis;
   dest->blank = src.blank;
   dest->useblank = src.useblank;
   dest->datamax = src.datamax;
   dest->datamin = src.datamin;
   strncpy(dest->object, src.object,80);
   /* Set the arrays.  If they already exist, free them first */
   if (dest->naxis != (int *) NULL) free(dest->naxis);
   if (dest->crval != (float *) NULL) free(dest->naxis);
   if (dest->cdelt != (float *) NULL) free(dest->naxis);
   if (dest->crpix != (float *) NULL) free(dest->naxis);
   dest->naxis = (int *) malloc ( dest->n_axis * sizeof(int));
   dest->crval = (float *) malloc ( dest->n_axis * sizeof(float));
   dest->cdelt = (float *) malloc ( dest->n_axis * sizeof(float));
   dest->crpix = (float *) malloc ( dest->n_axis * sizeof(float));
   dest->cd = (float *) malloc ( dest->n_axis*dest->n_axis * sizeof(float));
   dest->pc = (float *) malloc ( dest->n_axis*dest->n_axis * sizeof(float));

   for (i = 0 ; i < src.n_axis ; ++i) {
      dest->naxis[i] = src.naxis[i];
      dest->crpix[i] = src.crpix[i];
      dest->cdelt[i] = src.cdelt[i];
      dest->crval[i] = src.crval[i];
   }
	for (i = 0 ; i < src.n_axis*src.n_axis; ++i) {
		dest->pc[i] = src.pc[i];
		dest->cd[i] = src.cd[i];
	}
   /* Free up any old other nodes */
   free_other_nodes(dest);
   /*dest->other = (struct other_node *) NULL;*/
   for(node = src.other ; node != (struct other_node *) NULL ; node=node->next){
      addkey(dest, node->key, node->eq, node->value);
   }

   return(0);
}

/* a routine to copy the data portion of a fits variable from src to dest. */
int cp_data( fits input, fits *output) {

   int i;
   long int ndata=1;

   /* assume the data area has not been allocated */
   if (output->data == (float *)NULL) alloc_fits(output);

   for (i = 0 ; i < input.n_axis ; ++i) {
      ndata = ndata*input.naxis[i];
   }

   for (i = 0 ; i < ndata ; ++i ) {
      output->data[i] = input.data[i];
   }
   return(0);
}


/*  Determine whether or not we need to swap bytes on a machine.  This should
    only be needed internally */
int must_swap()            /* check on byte/word swapping needs */
{
   unsigned long lw = 0x0A0B0C0D;        /* any test value */
   unsigned long hton_lw = htonl(lw);        /* go to Net-order = MSB first */
   if( hton_lw != lw ) return(1);        /* Must byte-swap the data */
   else return(0);                        /* no change ==> no swap */
}

void l_swap(unsigned char *lword)                 /* swap a 4 byte word */
{
   unsigned char tmp;
   tmp = lword[0]; lword[0] = lword[3]; lword[3] = tmp;
   tmp = lword[1]; lword[1] = lword[2]; lword[2] = tmp;
}

void s_swap(unsigned char *sword)                  /* swap a 2 byte word */
{
   unsigned char tmp;
   tmp = sword[0]; sword[0] = sword[1]; sword[1] = tmp;
}

/* allocate memory for the data portion of the fits struct */
int alloc_fits ( fits *var ) {
   int i;
   long int ndata=1;
   int size;

   size = abs(var->bitpix);
   if(fdebug) fprintf(stderr, "size = %d\n", size);

   for (i = 0 ; i < var->n_axis ; ++i) {
      ndata = ndata*var->naxis[i];
   }
   if(fdebug) fprintf(stderr, "ndata = %ld\n",ndata);
   if ( ! ( var->data = (float *) malloc(ndata*size))) {
      return(-1);
   }
   else {
      return(0);
   }
}

/*  Routine to read in the data after the file pointed has been advanced to 
    the beginning of the data segment (i.e., read_header must be called first)
*/
void read_data(FILE *fp, fits *var) 
{
   int k;
   int ndata=1;
   unsigned char cpixel;
   /*unsigned short int spixel;*/
   short int spixel;
   unsigned long int lpixel;
   float fpixel;
   int swap_bytes = must_swap();     /* NB: FITS is always MSB/NET ordered */

   for ( k = 0 ; k < var->n_axis ; ++k ) {
      ndata *= var->naxis[k];
   }

   if( var->bitpix == 8 ) {
      for( k = 0 ; k < ndata ; ++k ) {
         fread(&cpixel,1,1,fp);
         var->data[k] = (float) cpixel;
      }
   }
   else
   if( var->bitpix == 16 ) {                        /* 2 byte integers */
      for( k = 0 ; k < ndata ; ++k ) {
         fread(&spixel,2,1,fp);
         if( swap_bytes ) s_swap((unsigned char *) &spixel);
         var->data[k] = (float) spixel;
      }
   }
   else
   if( var->bitpix == 32 ) {                        /* 4 byte integers */
      for( k = 0 ; k < ndata ; ++k ) {
         fread(&lpixel,4,1,fp);
         if( swap_bytes ) l_swap((unsigned char *) &lpixel);
         var->data[k] = (float) lpixel;
      }
   }
   else
   if( var->bitpix == -32 ) {                        /* 4 byte floats */
      for( k = 0 ; k < ndata ; ++k ) {
         fread(&fpixel,4,1,fp);
         if( swap_bytes ) l_swap((unsigned char *) &fpixel);
         var->data[k] = (float) fpixel;
      }
   }
}

/* Read a single pixel value at coordinates (i).  Remember, coordinates
   run from 0..(naxis - 1).  The function rewinds the *fp, reads the header
   and then offsets and reads the value at the specified location */
/* Because I've gone to arbitrary n_axis, we must use the absolute off
 * set rather than specifying a coordinate tuple */
float read_fits_pixel (FILE *fp, int i) {
   fits dummy;
   char cpixel;
   short int spixel;
   long int lpixel;
   float fpixel;
   unsigned long int index;
   float value;
   int swap_bytes = must_swap();     /* NB: FITS is always MSB/NET ordered */

   /* first off, let's make sure we are at the beginning of the file stream */
   rewind(fp);

   /* Next, read the header, so that we are at the head of the data stream */
   read_header(fp, &dummy);

   /* Now fseek to the proper location */
   index = i*abs(dummy.bitpix)/8;
   fseek(fp, index, SEEK_CUR);

   if( dummy.bitpix == 8 ) {
      fread(&cpixel,1,1,fp);
      value = (float) cpixel;
   }
   else
   if( dummy.bitpix == 16 ) {                        /* 2 byte integers */
      fread(&spixel,2,1,fp);
      if( swap_bytes ) s_swap((unsigned char *) &spixel);
      value = (float) spixel;
   }
   else
   if( dummy.bitpix == 32 ) {                        /* 4 byte integers */
      fread(&lpixel,4,1,fp);
      if( swap_bytes ) l_swap((unsigned char *) &lpixel);
      value = (float) lpixel;
   }
   else
   if( dummy.bitpix == -32 ) {                        /* 4 byte floats */
      fread(&fpixel,4,1,fp);
      if( swap_bytes ) l_swap((unsigned char *) &fpixel);
      value = (float) fpixel;
   }
   return(value);
}

/* Write a fits file to the named file.  A wrapper for write_header and 
   write_data */
int write_fits_file(char *filename, fits var) {
   int result;
   FILE *fp;

   if ( ! ( fp = fopen(filename , "w") ) ) {
      fprintf(stderr, "Could not write to file %s\n", filename);
      return(-1);
   }

   result = write_header(fp, var);
   if(result < 0) return result;
   result = write_data(fp, var);
   if(result < 0) return result;

   fclose(fp);
   return(0);
}


int write_data(FILE *fp, fits var)
{
    int k;
    short int spixel;
    long int lpixel, ndata=1;
    float fpixel;
    char cpixel;
    char wpixel = 0;
    short bytes = 0;
    int swap_bytes = must_swap();     /* NB: FITS is always MSB/NET ordered */

    for( k = 0 ; k < var.n_axis ; ++k ) {
       ndata = ndata*var.naxis[k];
       fprintf(stderr, "naxis[%d] = %d\n", k, var.naxis[k]);
    }
    fprintf(stderr, "bitpix = %d, ndata = %ld\n", var.bitpix, ndata);
    if( var.bitpix == 8 ) {
       bytes = 1;
       for( k = 0 ; k < ndata ; ++k) {
          cpixel = (char) var.data[k];
          fwrite(&cpixel,1,1,fp);
       }
    }
    else
    if( var.bitpix == 16 ) {                        /* 2 byte integers */
       bytes = 2;
       for( k = 0 ; k < ndata ; ++k ) {
          spixel = (short int) var.data[k];
          if( swap_bytes ) s_swap((unsigned char *) &spixel);
          fwrite(&spixel,2,1,fp);
       }
    }
    else
    if( var.bitpix == 32 ) {                        /* 4 byte integers */
       bytes = 4;
       for( k = 0 ; k < ndata ; ++k ) {
          lpixel = (long int) var.data[k];
          if( swap_bytes ) l_swap((unsigned char *) &lpixel);
          fwrite(&lpixel,4,1,fp);
       }
    }
    else
    if( var.bitpix == -32 ) {                        /* 4 byte floats */
       bytes = 4;
       for( k = 0 ; k < ndata ; ++k ) {
          fpixel = (float) var.data[k];
          if( swap_bytes ) l_swap((unsigned char *) &fpixel);
          fwrite(&fpixel,4,1,fp);
       }
    }
    for ( k = (bytes*ndata) % 2880 ; k < 2880 ; ++k) {
       fwrite(&wpixel, 1, 1, fp);
    }
    return(k);
}

/* Read in a fits file.  A wrapper for read_fits */
int read_fits_file ( char *filename, fits *var) {
   int result;
   FILE *fp;

   if ( ! (fp = fopen(filename, "r")) ) {
      fprintf(stderr, "Could not open file %s\n", filename);
      return(1);
   }

   result = read_fits( fp, var);
   fclose(fp);
   return(result);
}


/* read_fits is the wrapper for read_header and read_data that handles internal
   calls */

int read_fits ( FILE *fp, fits *var ) {

   int result;

   /* First and foremost, if there is any pre-existing data, free it up: */
   free_other_nodes(var);
   if (var->data == (float *)NULL) {
      free(var->data);
      var->data = (float *)NULL;
   }

   /* First off, read the header */
   result = read_header(fp, var);
   if (fdebug) print_header(*var);
   if (result < 0) {
      return(-1);     /* There was an error */
   }

   /* Then, allocate the memory for the data */
   result = alloc_fits(var);
   if ( result != 0 ) {
      return(-1);     /* There was an error */
   }

   /* Finally, read in the data */
   read_data(fp, var);

   return(0);
}

/* Compute the size of the data */
long int get_ndata( fits var)
{
   int i=0;
   long int ndata=1;
   for (i = 0; i < var.n_axis ; ++i) ndata = ndata*var.naxis[i];
   return ndata;
}

/*  READ out the fits header  */
void print_header( fits var )
{
   int i;

   fprintf(stderr,"SIMPLE = %d  BITPIX=%d  NAXIS=%d\n", var.simple, var.bitpix, 
           var.n_axis);
   fprintf(stderr,"DATAMAX=%f    DATAMIN=%f \n", var.datamax, var.datamin);
   for (i = 0 ; i < var.n_axis ; ++i) {
      fprintf(stderr,"NAXIS%d=%d   CRPIX%d=%f   CRVAL%d=%f   CDELT%d=%f\n",i+1,
              var.naxis[i], i+1, var.crpix[i], i+1, var.crval[i], i+1,
              var.cdelt[i]);
   }
}

char * get_tag(fits *self,char * tag) {
   char *contents;
   int result;
   contents = malloc(80*sizeof(char));
   if(strcasecmp(tag,"simple") == 0) {
      sprintf(contents, "%d", self->simple);
   }
   else if(strcasecmp(tag,"bitpix") == 0){
      sprintf(contents, "%d", self->bitpix);
   }
   else if(strcasecmp(tag,"naxis") == 0) {
      sprintf(contents, "%d", self->n_axis);
   }
   else if(strcasecmp(tag,"blank") == 0){
      sprintf(contents, "%x", self->blank);
   }
   else if(strcasecmp(tag,"datamax") == 0){
      sprintf(contents, "%f", self->datamax);
   }
   else if(strcasecmp(tag,"datamin") == 0){
      sprintf(contents, "%f", self->datamin);
   }
   else if(strcasecmp(tag,"object") == 0){
      sprintf(contents, "%s", self->object);
   }
   else {
      result = search_key( *self, tag, contents);
      if ( result != 0 ) {
         sprintf(contents, "N/A");
      }
   }
   return(contents);
}
#ifdef __cplusplus
}
#endif
